from .network import NetworkConfig
from .database import DatabaseConfig
from .structute import StructureConfig
from .drivers import DriversConfig
from .docs import DocsConfig
from .dev import DevConfig



import typing

from jija import views

from aiohttp import web


class Router:
    def __init__(self, endpoints):
        self.__endpoints = endpoints
        self.__routes = self.__generate_routes(endpoints)

    @property
    def routes(self):
        return self.__routes

    @property
    def endpoints(self):
        return self.__endpoints

    @staticmethod
    def __generate_routes(endpoints):
        result = []
        for endpoint in endpoints:
            result.extend(endpoint.generate_routes())

        return result

    def __repr__(self):
        return f'<{self.__class__.__name__}: {len(self.endpoints)} endpoints, {len(self.__routes)} routes>'

    def __str__(self):
        return self.__repr__()


class AbsEndpoint:
    def generate_routes(self, prefix=''):
        raise NotImplementedError()


class Endpoint(AbsEndpoint):
    def __init__(self, path, view: typing.Type[views.ViewBase]):
        if not issubclass(view, views.ViewBase):
            raise AttributeError(f'view must be a subclass of "jija.views.ViewBase", got {view}')

        self.__path = path
        self.__view = view

    @property
    def path(self) -> str:
        return self.__path

    @property
    def view(self) -> typing.Type[views.View]:
        return self.__view

    def generate_routes(self, prefix=''):
        result = []
        for method in self.__view.get_methods():
            result.append(web.route(method, f'{prefix}{self.__path}', self.__view.construct))

        return result


class Include(AbsEndpoint):
    def __init__(self, path, endpoints):
        if path[-1] == '/':
            raise ValueError('Include path must ends without "/"')

        self.__path = path
        self.__endpoints = endpoints

    @property
    def path(self):
        return self.__path

    @property
    def endpoints(self):
        return self.__endpoints

    def generate_routes(self, prefix=''):
        result = []
        for endpoint in self.__endpoints:
            result.extend(endpoint.generate_routes(f'{self.__path}{prefix}'))

        return result



from jija import docs
from . import base, fields


class DocsConfig(base.Config):
    URL: str = fields.CharField(default='/docs')

    def __init__(self, url=None):
        super().__init__(url=url)

    @classmethod
    def base_app_update(cls, aiohttp_app):
        docs_setuper = docs.DocsSetup(aiohttp_app)
        aiohttp_app = docs_setuper.setup()
        return aiohttp_app



import inspect

from typing import Dict

from jija.serializers import fields
from jija.serializers.exceptions import ValidationError


class Serializer:
    def __init__(self, data=None):
        self.data = data or {}
        self.errors = {}

        self.__valid = None
        self.__fields = self.get_fields()

    @classmethod
    def get_fields(cls) -> Dict[str, fields.Field]:
        serializer_fields = {}

        for name, obj in cls.__dict__.items():
            if issubclass(type(obj), fields.Field):
                serializer_fields[name] = obj

        return serializer_fields

    async def in_serialize(self):
        for name, field in self.__fields.items():
            try:
                self.data[name] = await field.validate(self.data.get(name))

            except ValidationError as exception:
                self.errors[name] = exception.error
                self.data[name] = exception.value

        if inspect.iscoroutinefunction(self.clean):
            await self.clean()
        else:
            self.clean()

        self.__valid = len(self.errors) == 0
        if not self.__valid:
            print(self.errors)

        return self.__valid

    async def out_validate(self):
        pass

    @property
    def valid(self):
        return self.__valid

    def clean(self):
        pass



from jija.serializers.fields import *
from jija.serializers import validators


class SubclassValidator(validators.Validator):
    @classmethod
    async def validate(cls, value, field):
        if not issubclass(value, field.class_pattern):
            raise ValidationError(f'Value must be subclass of {field.class_pattern}, not {type(value)}', value)

        return value
    

class ClassField(Field):
    validators = [SubclassValidator]

    def __init__(self, *, class_pattern, **kwargs):
        self.class_pattern = class_pattern
        super().__init__(**kwargs)



import asyncio
import inspect


class Command:
    def __init__(self):
        self.loop = asyncio.get_event_loop()

    async def prepare(self):
        pass

    def run(self):
        asyncio.get_event_loop().run_until_complete(self.prepare())

        if inspect.iscoroutinefunction(self.handle):
            self.loop.run_until_complete(self.handle())
        else:
            self.handle()

    def handle(self):
        raise NotImplementedError()



import inspect
import json
from typing import Union

import aiohttp.http_websocket
from aiohttp import web

from jija import response, serializers, exceptions


class ViewBase:
    def __init__(self, request: web.Request, path_params: web.UrlMappingMatchInfo):
        self.__request = request
        self.__path_params = path_params

    @property
    def request(self) -> web.Request:
        return self.__request

    @classmethod
    def get_methods(cls):
        raise NotImplementedError()

    @classmethod
    async def construct(cls, request: web.Request):
        view = cls(request, request.match_info)
        return await view.wrapper()

    async def wrapper(self):
        try:
            return await self.dispatch()

        except exceptions.ViewForceExit as exception:
            return exception.response

    async def dispatch(self):
        raise NotImplementedError()


class View(ViewBase):
    methods = ('get', 'post', 'patch', 'put', 'delete')

    serializers_in = None

    # serializers_out = None TODO

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.__method = self.request.method.lower()
        self.__data: dict = {}

    @property
    def method(self) -> str:
        return self.__method

    @property
    def data(self) -> dict:
        return self.__data

    @classmethod
    def get_methods(cls):
        view_methods = []
        for method in cls.methods:
            if hasattr(cls, method):
                view_methods.append(method)

        return view_methods

    async def dispatch(self):
        try:
            await self.load_data()
            handler = getattr(self, self.method)
            return await handler() if inspect.iscoroutinefunction(handler) else handler()

        except serializers.SerializeError as error:
            return response.JsonResponse(error.serializer.errors, status=400)

    async def load_data(self):
        data = {
            **await self.parse_body(),
            **self.parse_path(),
            **self.parse_query()
        }

        self.__data = await self.in_serialize(data)

    async def parse_body(self) -> dict:
        if self.method != 'get':
            try:
                return await self.request.json()
            except json.JSONDecodeError:
                return {}

        return {}

    def parse_path(self) -> dict:
        return dict(self.request.match_info)

    def parse_query(self) -> dict:
        data = {}

        for key in set(self.request.query.keys()):
            value = self.request.query.getall(key)
            if len(value) == 1:
                value = value[0]

            data[key] = value

        return data

    async def in_serialize(self, data: dict) -> dict:
        serializer_class = await self.get_in_serializer(self.method)
        if serializer_class:
            serializer = serializer_class(data)
            await serializer.in_serialize()

            if not serializer.valid:
                raise serializers.SerializeError(serializer)

            return serializer.data

        return data

    async def out_serialize(self, response) -> web.Response:
        # if isinstance(response, SerializeResponse): TODO
        #
        #     serializer_class = await self.get_out_serializer(self.method)
        #     if not serializer_class:
        #         raise ValueError('Got SerializerResponse, but out serializer not set')
        #
        #     return await response.serialize(serializer_class)

        return response

    @classmethod
    async def get_in_serializer(cls, method):
        return cls.serializers_in and cls.serializers_in[method]

    # @classmethod TODO
    # async def get_out_serializer(cls, method):
    #     return cls.serializers_out and cls.serializers_out[method]


class SerializersSet:
    def __init__(self, get=None, post=None, put=None, path=None, delete=None, **kwargs):
        self.__serializers = {
            'get': get,
            'post': post,
            'put': put,
            'path': path,
            'delete': delete,
            **kwargs
        }

    def __getitem__(self, item):
        return self.__serializers[item]


class DocMixin:
    pass


class WSView(ViewBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__ws = None

    @classmethod
    def get_methods(cls):
        return 'get',

    @property
    def ws(self) -> web.WebSocketResponse:
        return self.__ws

    async def dispatch(self):
        await self.on_connect()
        await self.process()
        return self.ws

    async def on_connect(self):
        self.__ws = web.WebSocketResponse()
        await self.ws.prepare(self.request)

    async def process(self):
        async for message in self.ws:
            message: aiohttp.http_websocket.WSMessage

            if message.type == web.WSMsgType.TEXT:
                await self.on_message(message.data)
            elif message.type == web.WSMsgType.ERROR:
                await self.on_error(message.data)

    async def on_message(self, message):
        pass

    async def on_error(self, error: str):
        await self.close(code=500, message=error, force=True)

    async def send(self, message: Union[dict, str, bytes]):
        if isinstance(message, dict):
            await self.ws.send_json(message)
        elif isinstance(message, bytes):
            await self.ws.send_bytes(message)
        else:
            await self.ws.send_str(message)

    async def close(self, code: int = 1000, message: Union[str, bytes, dict] = None, force=False):
        if isinstance(message, dict):
            message = json.dumps(message).encode('utf-8')
        elif isinstance(message, str):
            message = message.encode('utf-8')

        await self.ws.close(code=code, message=message)
        if force is True:
            raise exceptions.ViewForceExit(self.ws)



import re
import typing

from aiohttp import web
from typing import List, Dict

from jija import views, serializers, router
from jija import apps, app


class DocsProcessor:
    def __init__(self, aiohttp_app: web.Application):
        self.__base_aiohttp_app = aiohttp_app

    async def create_json(self):
        paths = await self.__parse_router('', apps.Apps.core)

        return {
            'openapi': '3.0.1',
            'paths': paths,
        }

    async def __parse_router(self, prefix, jija_app: app.App) -> dict:
        paths = {}

        router_views = self.get_router_views(jija_app.router.endpoints, prefix)

        for path, view in router_views.items():
            path_params = self.parse_path_params(path)
            doc_view = DocView(jija_app.name, view, path_params)
            paths[path] = await doc_view.parse()

        for sub_app in jija_app.childes:
            next_prefix = f'{prefix}{sub_app.get_url_prefix()}'
            paths.update(await self.__parse_router(next_prefix, sub_app))

        return paths

    def get_router_views(
            self,
            endpoints: List[typing.Union[router.Endpoint, router.Include]],
            prefix: str = ''
    ) -> dict:

        router_views = {}
        for endpoint in endpoints:
            if isinstance(endpoint, router.Endpoint):
                if issubclass(endpoint.view, views.DocMixin):
                    router_views[f'{prefix}{endpoint.path}'] = endpoint.view

            elif isinstance(endpoint, router.Include):
                router_views.update(self.get_router_views(endpoint.endpoints, f'{prefix}{endpoint.path}'))

        return router_views

    @staticmethod
    def parse_path_params(path: str) -> Dict[str, serializers.fields.Field]:
        result = {}
        for param in re.findall(r'{(\w*)}', path):
            result[param] = serializers.fields.CharField(required=True)

        return result


class DocView:
    def __init__(
            self,
            app_name,
            view: typing.Type[typing.Union[views.View, views.DocMixin]],
            path_params: Dict[str, serializers.fields.Field]
    ):
        self.__app_name = app_name
        self.__view = view
        self.__path_params = path_params

    async def parse(self):
        methods = {}
        for method in self.__view.get_methods():
            methods[method] = {
                'summary': getattr(self.__view, method).__doc__,
                'tags': [f'{self.__app_name} {self.__view.__name__}'],
                'parameters': self.create_params(self.__path_params, 'path'),

                "responses": {
                    "default": {}
                }
            }

            serializer_in = await self.__view.get_in_serializer(method)

            if serializer_in:
                if method == 'get':
                    methods[method]['parameters'] += self.create_params(serializer_in.get_fields(), 'query')

                else:
                    methods[method]["requestBody"] = self.__parse_serializer_in(serializer_in)

        return methods

    @staticmethod
    def __parse_serializer_in_get(serializer: serializers.Serializer):
        fields = []
        for name, field in serializer.get_fields().items():
            fields.append({
                'name': name,
                'in': 'query',
                'required': field.required,
                "schema": field.doc_get_schema()
            })

        return fields

    @staticmethod
    def __parse_serializer_in(serializer: serializers.Serializer):
        fields = serializer.get_fields()

        required = []
        properties = {}
        for name, field in fields.items():
            if field.required:
                required.append(name)

            properties[name] = field.doc_get_schema()

        return {
                "content": {
                    "application/json": {
                        "schema": {
                            "type": "object",
                            "required": required,
                            "properties": properties
                        }
                    }
                }
            }

    @staticmethod
    def create_params(params: Dict[str, serializers.fields.Field], placement: str) -> List[dict]:
        result = []
        for name, field in params.items():
            result.append({
                'name': name,
                'in': placement,
                'required': field.required,
                "schema": field.doc_get_schema()
            })

        return result



import os
from jija import config
from aiohttp_swagger import _swagger_home
import aiohttp_swagger
from . import views


class DocsSetup:
    STATIC_PATH = os.path.abspath(os.path.join(os.path.dirname(aiohttp_swagger.__file__), "swagger_ui3"))

    def __init__(self, aiohttp_app):
        self.__aiohttp_app = aiohttp_app

    def setup(self):
        self.__aiohttp_app.router.add_route('GET', f'{config.DocsConfig.URL}/', _swagger_home)
        self.__aiohttp_app.router.add_route(
            'GET', f"{config.DocsConfig.URL.rstrip('/')}/swagger.json", views.swagger_view)

        static_route = f'{config.DocsConfig.URL}/swagger_static'

        self.__aiohttp_app.router.add_static(static_route, self.STATIC_PATH)

        self.__aiohttp_app["SWAGGER_DEF_CONTENT"] = 'asdasd'
        self.__set_template()
        return self.__aiohttp_app

    def __set_template(self):
        static_route = f'{config.DocsConfig.URL}/swagger_static'

        with open(os.path.join(self.STATIC_PATH, "index.html"), "r") as file:
            index_html = file.read()

        self.__aiohttp_app["SWAGGER_TEMPLATE_CONTENT"] = (
            index_html
            .replace("##SWAGGER_CONFIG##", f"{config.DocsConfig.URL.rstrip('/')}/swagger.json")
            .replace("##STATIC_PATH##", f"{static_route}")
            .replace("##SWAGGER_VALIDATOR_URL##", "")
        )



from aiohttp import web

from jija import response
from . import processor


async def swagger_view(request):
    doc_processor = processor.DocsProcessor(request.app)
    return response.JsonResponse(data=await doc_processor.create_json())



from jija.serializers.validators import *
from jija.serializers.exceptions import ValidationError


class Field:
    DOC_TYPE = NotImplemented
    DOC_FORMAT = None

    validators = ()

    def __init__(self, *, required=True, default=None):
        self.required = required
        self.default = default

    async def validate(self, value):
        if value is None:
            if self.required and self.default is None:
                raise ValidationError('Required field', value)
            else:
                return value or self.default

        for validator in self.validators:
            value = await validator.validate(value, self)

        return value

    def doc_get_schema(self):
        return {
            **self.doc_get_type_data(),
            **self.doc_get_extra()
        }

    def doc_get_type_data(self):
        type_data = {
            'type': self.DOC_TYPE
        }

        if self.DOC_FORMAT:
            type_data['format'] = self.DOC_FORMAT

        return type_data

    def doc_get_extra(self):
        return {}


class CharField(Field):
    DOC_TYPE = 'string'

    validators = (LengthMinValidator, LengthMaxValidator)

    def __init__(self, *, min_length=None, max_length=None, regex=None, **kwargs):
        super().__init__(**kwargs)
        self.min_length = min_length
        self.max_length = max_length
        self.regex = regex

    def doc_get_extra(self):
        extra = {}
        if self.max_length is not None:
            extra['maxLength'] = self.max_length

        if self.min_length is not None:
            extra['minLength'] = self.min_length

        return extra


class NumericField(Field):
    def __init__(self, *, min_value=None, max_value=None, **kwargs):
        super().__init__(**kwargs)
        self.min_value = min_value
        self.max_value = max_value

    def doc_get_extra(self):
        extra = {}

        if self.max_value is not None:
            extra['maximum'] = self.max_value

        if self.min_value is not None:
            extra['minimum'] = self.min_value

        return extra


class IntegerField(NumericField):
    DOC_TYPE = 'integer'
    DOC_FORMAT = 'int32'

    validators = (IntegerValidator, RangeMinValidator, RangeMaxValidator)


class FloatField(NumericField):
    DOC_TYPE = 'number'
    DOC_FORMAT = 'float'

    validators = (FloatValidator, RangeMinValidator, RangeMaxValidator)


class DateField(Field):
    DOC_TYPE = 'string'
    DOC_FORMAT = 'date'

    validators = (DateValidator, RangeMinValidator, RangeMaxValidator)

    def __init__(self, *, min_value=None, max_value=None, **kwargs):
        super().__init__(**kwargs)
        self.min_value = min_value
        self.max_value = max_value


class SelectField(Field):
    validators = (OptionsValidator,)

    def __init__(self, *, options, **kwargs):
        super().__init__(**kwargs)
        self.options = options



from jija.drivers import base


class DatabaseDriver(base.Driver):
    @classmethod
    async def preflight(cls):
        raise NotImplementedError()

    @classmethod
    async def get_connection(cls):
        raise NotImplementedError()


class JijaOrmDriver(DatabaseDriver):
    @classmethod
    async def preflight(cls):
        from jija_orm import config as jija_orm_config
        from jija import config

        await jija_orm_config.JijaORM.async_init(
            project_dir=config.StructureConfig.PROJECT_PATH,
            connection=jija_orm_config.Connection(
                host=config.DatabaseConfig.HOST,
                port=config.DatabaseConfig.PORT,
                user=config.DatabaseConfig.USER,
                database=config.DatabaseConfig.DATABASE,
                password=config.DatabaseConfig.PASSWORD
            ),
            apps=await cls.__get_apps()
        )

    @classmethod
    async def __get_apps(cls):
        from jija_orm import config as jija_orm_config
        from jija.apps import Apps

        apps = []
        for app in Apps.apps.values():
            if app.exist('models.py'):
                path = app.name if app.parent is None else f'apps.{app.name}'
                apps.append(jija_orm_config.App(
                    name=app.name, path=path, migration_dir=app.get_import_path('migrations')))

        return apps

    @classmethod
    async def get_connection(cls):
        pass




from . import config
from . import commands
from . import contrib
from . import docs
from . import drivers
from . import middlewares
from . import serializers
from . import app
from . import apps
from . import collector
from . import command
from . import exceptions
from . import reloader
from . import response
from . import router
from . import views



class ValidationError(Exception):
    def __init__(self, error, value):
        self.error = error
        self.value = value


class SerializeError(Exception):
    def __init__(self, serializer):
        self.__serializer = serializer

    @property
    def serializer(self):
        return self.__serializer



from aiohttp import web
from jija.command import Command


class RunProcess(Command):
    def handle(self):
        from jija.apps import Apps
        from jija.config import NetworkConfig

        web.run_app(Apps.core.aiohttp_app, loop=self.loop, host=NetworkConfig.HOST, port=NetworkConfig.PORT)



import asyncio
import signal
import subprocess
import sys

from jija import config
from jija.command import Command
from jija import reloader


class Run(Command):
    def __init__(self):
        super().__init__()
        self.close_event = asyncio.Event()
        self.reloader = reloader.Reloader(config.StructureConfig.PROJECT_PATH, self.close_event)
        self.runner = None
        self.alive = True

        signal.signal(signal.SIGTERM, self.signal_handler)

    def signal_handler(self, sig, frame):
        self.alive = False
        self.close_event.set()

    async def run_watcher(self):
        await self.reloader.wait()

    async def handle(self):
        asyncio.create_task(self.run_watcher())
        while self.alive:
            self.runner = subprocess.Popen([config.StructureConfig.PYTHON_PATH, 'main.py', 'runprocess'])
            self.close_event.clear()
            await self.close_event.wait()

            self.close_sub_process()
            print()

    def close_sub_process(self):
        if sys.platform == "win32":
            self.runner.send_signal(signal.SIGTERM)
        else:
            self.runner.send_signal(signal.SIGINT)

    def run(self):
        try:
            super().run()
        except KeyboardInterrupt:
            self.reloader.close()
            self.close_sub_process()
            self.runner.wait()



import importlib
import os
from pathlib import Path
from typing import Optional

from aiohttp import web
from jija import config

from jija.middleware import Middleware
from jija.collector import collect_subclasses
from jija.command import Command
from jija import router


class App:
    ROUTER = router.Router
    CUSTOM_URL_PATH: Optional[str] = None

    def __init__(self, *, name, path, aiohttp_app=None, parent=None):
        """
        :type name: str
        :type path: jija.utils.path.Path
        :type aiohttp_app: web.Application
        :type parent: App
        """

        self.__parent = parent
        if parent:
            parent.add_child(self)

        self.__path = path
        self.__name = name
        self.__is_core = aiohttp_app is not None

        self.__router = None
        self.__middlewares = None
        self.__commands = None

        self.__aiohttp_app = None

        self.__childes = []

        self.__load(aiohttp_app)

    @property
    def parent(self):
        """:rtype: App"""
        return self.__parent

    @property
    def name(self):
        """:rtype: str"""
        return self.__name

    @property
    def router(self) -> router.Router:
        return self.__router

    @property
    def middlewares(self):
        """:rtype: list[Middleware]"""
        return self.__middlewares

    @property
    def aiohttp_app(self):
        """:rtype: web.Application"""
        return self.__aiohttp_app

    @property
    def childes(self):
        """:rtype: list[App]"""
        return self.__childes

    @property
    def path(self) -> Path:
        return self.__path

    @property
    def commands(self) -> list:
        return self.__commands

    def __load(self, aiohttp_app: web.Application = None):
        self.__router = self.get_router()
        self.__middlewares = self.__get_middlewares()
        self.__commands = self.__get_commands()

        self.__aiohttp_app = self.get_aiohttp_app(aiohttp_app)

    def get_router(self) -> "router.Router":
        if not self.exist('routes'):
            raw_routes = []

        else:
            import_path = self.get_import_path('routes')
            routes_module = importlib.import_module(import_path)

            raw_routes = getattr(routes_module, 'routes', [])

        if self.parent is None and self.CUSTOM_URL_PATH is not None:
            raw_routes = [router.Include(self.CUSTOM_URL_PATH, raw_routes)]

        app_router = self.ROUTER(raw_routes)
        return app_router

    def __get_middlewares(self) -> list:
        if not self.exist('middlewares'):
            return []

        import_path = self.get_import_path('middlewares')
        middlewares_module = importlib.import_module(import_path)
        middlewares = collect_subclasses(middlewares_module, Middleware)

        return list(map(lambda middleware: middleware(), middlewares))

    def __get_commands(self) -> dict:
        if not self.exist('commands'):
            return {}

        commands = {}
        commands_path = self.__path.joinpath('commands')
        for file in os.listdir(commands_path):
            if file.endswith('.py') and not file.startswith('_'):

                import_path = self.get_import_path(f'commands.{file.removesuffix(".py")}')
                command_module = importlib.import_module(import_path)

                command = list(collect_subclasses(command_module, Command))
                if command:
                    commands[file.replace('.py', '')] = command[0]

        return commands

    @staticmethod
    def is_app(path: Path) -> bool:
        if path.is_dir() is False or path.joinpath('app.py').exists() is False:
            return False

        for part in path.parts:
            if part.startswith('__'):
                return False

        return True

    def get_aiohttp_app(self, aiohttp_app: web.Application = None) -> web.Application:
        aiohttp_app = aiohttp_app or web.Application()

        aiohttp_app.middlewares.extend(self.__middlewares)

        aiohttp_app.add_routes(self.__router.routes)
        aiohttp_app['JIJA_ROUTER'] = self.__router

        return aiohttp_app

    def add_child(self, child: "App"):
        self.__childes.append(child)

    def get_import_path(self, to: str) -> str:
        modify_class_path = self.path.joinpath(to)
        return ".".join(modify_class_path.relative_to(config.StructureConfig.PROJECT_PATH).parts)

    def exist(self, name: str) -> bool:
        return self.__path.joinpath(name).exists() or self.__path.joinpath(f'{name}.py').exists()

    def register(self):
        for child in self.childes:
            child.register()

            self.aiohttp_app.add_subapp(prefix=child.get_url_prefix(), subapp=child.aiohttp_app)

    def get_url_prefix(self) -> str:
        """
        Method returns app path prefix
        If app is core we returns name or CUSTOM_URL_PATH
        If app is secondary we need to add prefix of first app
            because in core app we make Include if CUSTOM_URL_PATH is defined and app doesn't know itself name
        If app is thirded we returns only it prefix
        """
        if not self.parent:
            return self.CUSTOM_URL_PATH or ''

        self_prefix = self.CUSTOM_URL_PATH or f'/{self.name}'

        if not self.parent.parent:
            parent_prefix = self.parent.get_url_prefix()
        else:
            parent_prefix = ''

        return f'{parent_prefix}{self_prefix}'



import typing
from aiohttp.helpers import sentinel
from aiohttp.typedefs import LooseHeaders
from aiohttp.web import json_response, Response

from jija import serializers


class JsonResponse:
    def __new__(
            cls,
            data: typing.Any = sentinel,
            *,
            text: typing.Optional[str] = None,
            body: typing.Optional[bytes] = None,
            status: int = 200,
            reason: typing.Optional[str] = None,
            headers: typing.Optional[LooseHeaders] = None,
            content_type: str = "application/json"
    ):

        return json_response(
            data=data,
            text=text,
            body=body,
            status=status,
            reason=reason,
            headers=headers,
            content_type=content_type,
        )


class SerializeResponse:
    def __init__(self, data: typing.Any, status: int = 200):
        self.__data = data
        self.__status = status

    @property
    def data(self):
        return self.__data

    @property
    def status(self):
        return self.__status

    async def serialize(self, serializer: serializers.Serializer) -> Response:
        serializer = serializer(data=self.data)
        serialized_data = serializer.serialize_out()

        return json_response(
            body=serialized_data,
            status=self.status,
        )



from jija.config import base


class DevConfig(base.Config):
    RELOADER_EXCLUDED_DIRS: set

    def __init__(self, *, reloader_excluded=None):
        self.__class__.__PREF = {
            'RELOADER_EXCLUDED_DIRS': reloader_excluded if reloader_excluded else set()
        }

        super().__init__()

    @classmethod
    async def freeze(cls):
        from jija import config

        validated_data = set(
            str(config.StructureConfig.PROJECT_PATH.joinpath(path).absolute())
            for path in cls.__PREF['RELOADER_EXCLUDED_DIRS']
        )

        cls.set_values({
            'RELOADER_EXCLUDED_DIRS': validated_data
        })
